function [p,pe] = fmm2du(z,m,ze,varargin)
%FMM2Du Fast (uniform) multipole method in 2 dimensions.
%   P = FMM2Du(Z,M) evaluates the potential P due to masses M at
%   Cartesian coordinates Z = X+iY. It is understood that the size of
%   Z matches that of M.
%
%   PE = FMM2Du(Z,M,ZE) evaluates the potential at the points ZE
%   instead.
%
%   [P,PE] = FMM2Du(Z,M,ZE) evaluates the potential both at Z and ZE at
%   the same time.
%
%   P = FMM2Du(Z,M,...) or [...] = FMM2Du(Z,M,ZE,...) allows options to 
%   be passed in the form of property/value-pairs. See the table below 
%   for available properties.
%
%   Property     Value/{Default}     Description
%   -----------------------------------------------------------------------
%   pot          {0} | 1             0: Logarithmic potential,
%                                    P = -M*LOG(ABS(Z)).
%                                    1: Harmonic (doublet) potential,
%                                    P = -M/Z.
%
%   tol          Scalar {1e-6}       Tolerance. The choice tol = 0 means a 
%                                    direct evaluation.
%
%   ndirect      Scalar integer {5}  Approximately the number of particles
%                                    in each box at the finest level.
%
%   smooth       {'dirac'} |         Near-field models, see below.
%                'rankine' |
%                'scully' |
%                'oseen'
%
%   xopt         Scalar {1e-3}       Distance to the extreme point of the
%                                    smoother.
%
%   cutoff       Scalar {2e-3}       Cutoff-radius of the smoother.
%
%   cont         {true} | false      Enforce continuity to match the near- 
%                                    and far-fields.
%
%   The precise definitions of the smoothers as defined via the properties 
%   {'smooth','xopt','cutoff','cont'} are quite complicated. Instead, 
%   reference implementations have been prepared. See the files LOGPOT.M 
%   and INVPOT.M for details.
%
%   Cautionary: there are some pitfalls related to the spatial 
%   discreatization which depends on the input data (Z and/or ZE):
%
%   -Efficiency deteriorates if the smallest bounding box containing
%   ZE is considerably larger than the smallest bounding box
%   containing Z.
%
%   -The cut-off radius of the smoother is respected by the algorithm. If 
%   this radius is large compared to the computational domain, then the 
%   overall efficiency deteriorates.
%
%   Example:
%     % N random masses
%     N = 20;
%     z = complex(rand(N,1),rand(N,1));
%     m = rand(N,1);
%
%     % mesh
%     [X,Y] = meshgrid(linspace(0,1,40));
%     e = complex(X,Y);
%
%     % direct evaluation of the Logarithmic potential with Oseen smoothing
%     p = fmm2du(z,m,e,'tol',0,'smooth','oseen');
%     contour(X,Y,p);
%     hold on, plot(real(z),imag(z),'ko');
%
%   See also LOGPOT, INVPOT.

% S. Engblom 2011-02-02 (Museum version)
% S. Engblom 2007-07-08 (Major revision)
% S. Engblom 2007-05-15 (Revision)
% S. Engblom 2006-07-10
% Based on the code 'fmm2dlp' by the same author.

% Notes:
%   The call P = FMM2Du(Z,M,Z,...) is interpreted as P = FMM2Du(Z,M,...) 
%   through comparision of pointers.
%
%   The syntax Y = FMM2Du(X) computes Y = EXPINT(X) very rapidly. This 
%   syntax will probably be removed in future versions.

error('.MEX-file not found on path.');
