function p = invpot(z,m,ze,smooth,xopt,cutoff,cont)
%INVPOT Evaluation of harmonic potential.
%   P = INVPOT(Z,M,ZE,SMOOTH,XOPT,CUTOFF,CONT) is a reference
%   implementation of the harmonic potential -M/Z with masses M at
%   coordinates Z evaluated at ZE using the smoother SMOOTH =
%   {'dirac'} | 'rankine' | 'scully' | 'oseen' with parameters XOPT
%   (default 1e-3), CUTOFF (default 2e-3) and CONT {'on'} | 'off'.
%
%   XOPT is the distance from the potential to the extreme point of
%   abs(P) and CUTOFF is the cut-off radius for the smoother. It is
%   required that XOPT < CUTOFF. If CONT is 'on', then the smoother is
%   scaled to ensure continuity at the cut-off.
%
%   Example:
%     z = 0; m = 1; % a single potential at z = 0
%     x = linspace(-3e-3,3e-3);
%     [X,Y] = meshgrid(x,x);
%     ze = complex(X,Y); zf = complex(x,x);
%     cc = 100:100:1000;
%     s = {'dirac' 'rankine' 'scully' 'oseen'};
%     for i = 1:size(s,2)
%       p{i} = invpot(z,m,ze,s{i});
%       figure(1), subplot(2,2,i);
%       contour(x,x,abs(p{i}),cc); axis square, title(s{i});
%       q{i} = invpot(z,m,zf,s{i});
%     end
%     figure(2);
%     xx = sign(x).*abs(zf);
%     plot(xx,abs(q{1}),'b',xx,abs(q{2}),'r', ...
%          xx,abs(q{3}),'g',xx,abs(q{4}),'k');
%     legend(s{:}); axis([xx([1 end]) 0 1500]);
%
%   See also FMM2D, LOGPOT.

% S. Engblom 2007-05-16 (Revision)
% P. Deglaire 2006-07-17 (Modification)
% S. Engblom 2006-07-14

% defaults
if nargin < 7
  cont = 'on';
  if nargin < 6
    cutoff = 2e-3;
    if nargin < 5
      xopt = 1e-3;
      if nargin < 4
        smooth = 'dirac';
        if nargin < 3
          ze = z;
        end
      end
    end
  end
end

if xopt >= cutoff, error('Illegal parameters.'); end

if isempty(ze), ze = z; end

N = numel(z);
NE = numel(ze);
p = zeros(size(ze));

switch smooth
 case 'dirac'
  for i = 1:N
    for j = 1:NE
      if ze(j) ~= z(i) % protect from exactly zero
        p(j) = p(j)-m(i)/(ze(j)-z(i));
      end
    end
  end

 case 'rankine'
  cutoff = xopt;
  for i = 1:N
    for j = 1:NE
      d = ze(j)-z(i);
      ad = abs(d);
      if ad >= cutoff
        p(j) = p(j)-m(i)/d;
      else
        p(j) = p(j)-m(i)*conj(d)/cutoff^2;
      end
    end
  end

 case 'scully'
  shape = xopt;
  if strcmp(cont,'on')
    scale = 1+(shape/cutoff)^2;
  else
    scale = 1;
  end
  for i = 1:N
    for j = 1:NE
      d = ze(j)-z(i);
      ad = abs(d);
      if ad >= cutoff
        p(j) = p(j)-m(i)/d;
      else
        p(j) = p(j)-m(i)*conj(d)/(shape^2+abs(d)^2)*scale;
      end
    end
  end

 case 'oseen'
  shape = 1.2564312086261696770/xopt^2;
  if strcmp(cont,'on')
    scale = -1/expm1(-shape*cutoff^2);
  else
    scale = 1;
  end
  for i = 1:N
    for j = 1:NE
      d = ze(j)-z(i);
      ad = abs(d);
      if ad >= cutoff
        p(j) = p(j)-m(i)/d;
      elseif d ~= 0 % protect from exactly zero
        p(j) = p(j)+m(i)/d*expm1(-shape*ad^2)*scale;
      end
    end
  end

 otherwise, error('Unknown smoother.');
end
