/* fmm.h */

/* S. Engblom 2007-07-08 (Major revision) */
/* S. Engblom 2007-06-19 */

#ifndef __fmm_h
#define __fmm_h

/* Multipole expansion stored in the form of a vectorized tree. First
   comes the root, then the children of the root and so on. A
   Cartesian ordering is used locally at each level.

   Let (ix,iy) be the (zero-offset) local coordinates at level l. Then
   the local linear index is given by j = ix+(iy<<l) = ix+iy*snb where
   snb = (1<<l) is sqrt(number of boxes at level l). For convenience, the 
   global linear index is given by i = lptr[l]+j.

   The local linear index of the parent can be found from
   (ix>>1)+((iy>>1)<<(l-1)). Similarly, the 0-child can be found
   according to (ix<<1)+(iy<<(l+2)).

   The children are ordered as follows:
      2 | 3
    y -----
      0 | 1
        x

   The coefficients are to be understood as
      mpexp(z) = coeff1[0]*log(z-z0)+
      coeff1[1]/(z-z0)+coeff1[2]/(z-z0)^2+coeff1[3]/(z-z0)^3+...
   or, depending on the stage of the algorithm,
      mpexp(z) = coeff2[0]+coeff2[1]*(z-z0)+coeff2[2]*(z-z0)^2+...
*/

typedef struct {
  dcmplx z0;                  /* the midpoint of the box */
  dcmplx *coeff1,*coeff2;     /* multipole coefficients */
} mpexp;

typedef struct {
  double xmin,xmax,ymin,ymax; /* bounding box */
  int pcoeff;                 /* number of coefficients offset by one */
  mpexp *root;                /* root of the tree */
  int nlevel;                 /* number of levels offset by one */
  int *lptr;                  /* pointer to levels in tree */
  int *ix,*ixptr;             /* permutation of input */
} MPexp;

/* high-level routines */
void MPexp_malloc(MPexp *this,int N,int p,int Ndirect,
		  double xmin,double xmax,double ymin,double ymax,
		  double rcutoff);
void MPexp_init(MPexp *this,int N,
		const double *x,const double *y,
		const double *mr,const double *mi,
		int pot);
void MPexp_smooth(int pot,SMOOTHER smooth,double xopt,bool cont,
		  double *cutoff,double *shape,double *scale);
void MPexp_eval2(MPexp *this,
		 double *pr,double *pi,
		 const double *x,const double *y,
		 const double *mr,const double *mi,int pot,
		 SMOOTHER smooth,double xopt,double cutoff,bool cont);
void MPexp_eval(MPexp *this,
		double *qr,double *qi,
		int NE,const double *ex,const double *ey,
		const double *x,const double *y,
		const double *mr,const double *mi,int pot,
		SMOOTHER smooth,double xopt,double cutoff,bool cont);
void MPexp_free(MPexp *this);

/* low-level routines */
void mpexp_init(mpexp *this,int p,
                const double x,const double y,
		const double mr,const double mi,
		int pot);
void mpexp_eval2(mpexp *this,int p,
		 double *pr,double *pi,
		 const double *x,const double *y,
		 const int *ix,int begin,int end);
void mpexp_eval(mpexp *this,int p,
                double *qr,double *qi,
		double ex,double ey);
void mpexp_directInteract2(double *pr,double *pi,
			   const double *zr,const double *zi,
			   const double *mr,const double *mi,
			   const int *ix,
			   int begin1,int end1,int begin2,int end2,
			   int pot,SMOOTHER smooth,
			   double cutoff,double shape,double scale);
void mpexp_directInteract(double *qr,double *qi,
			  double er,double ei,
			  const double *zr,const double *zi,
			  const double *mr,const double *mi,
			  const int *ix,int begin,int end,
			  int pot,SMOOTHER smooth,
			  double cutoff,double shape,double scale);

#include "fmmshift.h" /*** clumsy! ***/

/*------------------------------------------------------------------------*/
void fmm2dInteract(int N,
		   const double *zr,const double *zi,
		   const double *mr,const double *mi,
		   int NE,
		   const double *er,const double *ei,
		   double *pr,double *pi,
		   double *qr,double *qi,
		   int pot,double tol,int Ndirect,
		   SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Evaluation by the fast multipole method to accuracy TOL of the
   potential from N pointmasses (mr,mi) (with mi possibly NULL) at the
   positions (zr,zi). The evaluation is done at the points (zr,zi) if
   (pr,pi) is not NULL and/or at the points (er,ei) if (qr,qi) is not
   NULL. The result is added to the vectors (pr,pi) and (qr,qi),
   respectively, which must be allocated prior to call. Logarithmic
   potential for pot == 0, harmonic potential otherwise. The smoothing
   is done using smoother smooth with parameters
   (xopt,cutoff,cont). */
{
  /* early return: cannot allocate anything */
  if (N <= 0) return;

  /* determine the bounding box */
  double xmin = zr[0],xmax = zr[0],ymin = zi[0],ymax = zi[0];
  for (int i = 1; i < N; i++) {
    if (zr[i] < xmin)
      xmin = zr[i];
    else if (zr[i] > xmax)
      xmax = zr[i];

    if (zi[i] < ymin)
      ymin = zi[i];
    else if (zi[i] > ymax)
      ymax = zi[i];
  }

  /* bounding box might become larger */
  if (qr != NULL)
    for (int i = 0; i < NE; i++) {
      if (er[i] < xmin)
        xmin = er[i];
      else if (er[i] > xmax)
        xmax = er[i];

      if (ei[i] < ymin)
        ymin = ei[i];
      else if (ei[i] > ymax)
        ymax = ei[i];
    }

  /* need quadratic computational domain */
  double hx = 0.5*(xmax-xmin),hy = 0.5*(ymax-ymin);
  double mx = 0.5*(xmax+xmin),my = 0.5*(ymax+ymin);
  if (hx >= hy) {
    ymin = my-hx;
    ymax = my+hx;
  }
  else {
    xmin = mx-hy;
    xmax = mx+hy;
  }

  /* determine the number of terms */
  int p = floor(-log2(tol))+(pot != 0);
  if (p < 1) p = 1;

  /* allocate multipole expansion */
  MPexp MP;
  if (smooth == DIRAC) cutoff = 0.0;
  MPexp_malloc(&MP,N,p,Ndirect,xmin,xmax,ymin,ymax,cutoff);

  /* initialize */
  MPexp_init(&MP,N,zr,zi,mr,mi,pot);

  /* evaluation at (zr,zi) */
  if (pr != NULL) MPexp_eval2(&MP,pr,pi,zr,zi,mr,mi,
			      pot,smooth,xopt,cutoff,cont);

  /* evaluation at (er,ei) */
  if (qr != NULL) MPexp_eval(&MP,qr,qi,NE,er,ei,zr,zi,mr,mi,
			     pot,smooth,xopt,cutoff,cont);

  /* dellocate */
  MPexp_free(&MP);
}
/*------------------------------------------------------------------------*/
void MPexp_malloc(MPexp *this,
		  int N,int p,int Ndirect,
		  double xmin,double xmax,double ymin,double ymax,
		  double cutoff)
/* Allocates a multipole tree for N potentials using p terms in the
   box [xmin,xmax] x [ymin,ymax]. The discretization is determined so
   that the number of boxes at the finest level is <~ N/Ndirect. Note,
   however, that the cutoff-radius is respected so that the near-field
   is correctly computed. */
{
  /* bounding box */
  this->xmin = xmin;
  this->xmax = xmax;
  this->ymin = ymin;
  this->ymax = ymax;

  /* number of terms */
  this->pcoeff = p;

  /* determine the discretization */
  this->nlevel = 0;
  int Nt = 1,Nf = 1;
  double hx = xmax-xmin,heff = hx; /* hy = hx by construction */
  for (this->nlevel = 0; Ndirect*Nf < N && cutoff < heff; this->nlevel++) {
    Nt += (Nf <<= 2);
    heff *= 0.5;
  }

  /* the while-part above should actually be true */
  if (this->nlevel > 0) {
    Nt -= Nf;
    Nf >>= 2;
    this->nlevel--;
  }

  /* multipole tree */
  this->root = mxMalloc(Nt*sizeof(mpexp));

  /* multipole coefficients */
  dcmplx *coeffs1 = mxCalloc((p+1)*Nt,sizeof(dcmplx));
  dcmplx *coeffs2 = mxCalloc((p+1)*Nt,sizeof(dcmplx));

  /* pointer to levels */
  this->lptr = mxMalloc((this->nlevel+2)*sizeof(int));

  /* first level */
  this->lptr[0] = 0;
  this->root[0].z0 = xmin+0.5*hx+I*(ymin+0.5*hx);
  this->root[0].coeff1 = coeffs1; coeffs1 += p+1;
  this->root[0].coeff2 = coeffs2; coeffs2 += p+1;
  this->lptr[1] = 1;
  hx *= 0.5;

  /* the rest of the levels: the incrementation logic ensures that the
     four sets of coefficients belonging to siblings are aligned in
     memory */
  for (int level = 1,i = 1,snb = 2; level <= this->nlevel; level++) {
    dcmplx Z0 = xmin+0.5*hx+I*(ymin+0.5*hx); /* lower left corner */
    for (int iy = 0; iy < snb; iy += 2) {
      dcmplx Z0_ = Z0; /* local copy to increment */
      for (int ix = 0; ix < snb; ix += 2) {

        /* process 4 siblings at a time */
        this->root[i].z0 = Z0_;
        this->root[i].coeff1 = coeffs1; coeffs1 += p+1;
        this->root[i].coeff2 = coeffs2; coeffs2 += p+1;
        this->root[i+1].z0 = Z0_+hx;
        this->root[i+1].coeff1 = coeffs1; coeffs1 += p+1;
        this->root[i+1].coeff2 = coeffs2; coeffs2 += p+1;
        this->root[i+snb].z0 = Z0_+I*hx;
        this->root[i+snb].coeff1 = coeffs1; coeffs1 += p+1;
        this->root[i+snb].coeff2 = coeffs2; coeffs2 += p+1;
        this->root[i+snb+1].z0 = Z0_+hx+I*hx;
        this->root[i+snb+1].coeff1 = coeffs1; coeffs1 += p+1;
        this->root[i+snb+1].coeff2 = coeffs2; coeffs2 += p+1;
        Z0_ += 2.0*hx;
        i += 2;
      }
      Z0 += 2.0*I*hx;
      i += snb;
    }

    /* pointer to next level */
    this->lptr[level+1] = i;

    /* new, finer level */
    snb <<= 1; /* sqrt(number of boxes on this level) */
    hx *= 0.5;
  }

  /* permutation vector */
  this->ix = mxMalloc(N*sizeof(int));
  this->ixptr = mxCalloc(Nf+1,sizeof(int));
}
/*------------------------------------------------------------------------*/
#define IMIN(a,b) ((a) <= (b) ? (a) : (b))

void MPexp_init(MPexp *this,int N,
		const double *x,const double *y,
		const double *mr,const double *mi,int pot)
/* Setup of the fast multipole method in three major steps. This
   function is critical to performance. */
{
  /* Step #1: Initialization of all potentials at the finest level. */
  {
    mpexp *offset = this->root+this->lptr[this->nlevel];
    const int snf = 1<<this->nlevel;
    const double hx = this->xmax-this->xmin; /* hy = hx by construction */
    int *ix = mxMalloc(N*sizeof(int));
    for (int ii = 0; ii < N; ii++) {
      /* determine position in tree */
      const double lx = (x[ii]-this->xmin)/hx,ly = (y[ii]-this->ymin)/hx;
      const int j = IMIN((int)floor(lx*snf),snf-1)+
      IMIN((int)floor(ly*snf),snf-1)*snf;

      /* initialize coefficients */
      mpexp_init(offset+j,this->pcoeff,
		 x[ii],y[ii],mr[ii],mi == NULL ? 0.0 : mi[ii],pot);

      /* remember where it went */
      ix[ii] = j;
      this->ixptr[j]++;
    }

    /* accumulate indices */
    for (int j = 1; j <= snf*snf; j++)
      this->ixptr[j] += this->ixptr[j-1];

    /* determine permutation of input */
    for (int ii = N-1; ii >= 0; ii--) 
      this->ix[--this->ixptr[ix[ii]]] = ii;
    mxFree(ix);
  }

  const int P = this->pcoeff+1;  /* used frequently below */
  shift_alloc(this->pcoeff,pot); /* prepare to shift */

  /* Step #2: Upward pass of the fast multipole algorithm. After the
     finest level has been initialized, this loop performs the
     'upward' pass where all multipole contributions are shifted and
     added onto all coarser levels. */
  for (int level = this->nlevel-1; level >= 0; level--) {
    mpexp *parent = &this->root[this->lptr[level]];
    mpexp *child = &this->root[this->lptr[level+1]];
    dcmplx *coeffs1 = child->coeff1;

    /* prepare for shift */
    shift_initLevel(child->z0-parent->z0,false);

    /* loop over all parents at this level... */
    while (parent < child) {
      /* ...shift and add result from each of 4 children */
      shift_m2m(parent->coeff1,coeffs1);
      parent++;
      coeffs1 += 4*P;
    }
  }

  /* Step #3: Downward pass of the fast multipole method. This shifts
     expansions from the interaction list of each box to the center of
     the box and adds up the result. Before continuing to finer
     levels, each parent shifts and adds its expansion to each of its
     children. */
  for (int level = 1,snb = 2; level < this->nlevel; level++, snb <<= 1) {
    mpexp *parent = &this->root[this->lptr[level]];
    mpexp *child = &this->root[this->lptr[level+1]];

    const dcmplx *coeffs1 = child->coeff1;
    dcmplx *coeffs2 = child->coeff2;
    const dcmplx *ucoeffs1 = coeffs1+4*(snb-1)*P;
    dcmplx *ucoeffs2 = coeffs2+4*(snb-1)*P;

    /* prepare for shift */
    shift_initLevel(parent->z0-child->z0,true);

    /* tricky part: loop over all boxes (parents) at this level and
       symmetrically shift results to and from the interaction list of
       each child */
    for (int iy = 0; iy < snb; iy++)
      for (int ix = 0; ix < snb; ix++) {

        /* parent to the right */
        if (ix < snb-1) {
          const dcmplx *rcoeffs1 = coeffs1+4*P;
          dcmplx *rcoeffs2 = coeffs2+4*P;

          shift_m2p(coeffs1,coeffs2,rcoeffs1,rcoeffs2,2,4);
          shift_m2p(coeffs1+P,coeffs2+P,rcoeffs1+P,rcoeffs2+P,2,1);
          shift_m2p(coeffs1+P,coeffs2+P,rcoeffs1+3*P,rcoeffs2+3*P,4,1);
          shift_m2p(coeffs1+2*P,coeffs2+2*P,rcoeffs1,rcoeffs2,0,4);
          shift_m2p(coeffs1+3*P,coeffs2+3*P,rcoeffs1+P,rcoeffs2+P,0,1);
          shift_m2p(coeffs1+3*P,coeffs2+3*P,rcoeffs1+3*P,rcoeffs2+3*P,2,1);
        }

        /* three parents above */
        if (iy < snb-1) {
          int m = 4,begin0 = 12,begin1 = 26,offset = 4;

          /* parent on the upper left */
          if (0 < ix) {
            m += 4;
            begin0 -= 4;
            begin1 -= 4;
            offset -= 4;

            shift_m2p(coeffs1+2*P,coeffs2+2*P,ucoeffs1,ucoeffs2,21,1);
            shift_m2p(coeffs1+2*P,coeffs2+2*P,ucoeffs1+2*P,ucoeffs2+2*P,8,2);
            shift_m2p(coeffs1+3*P,coeffs2+3*P,ucoeffs1,ucoeffs2,20,4);
          }

          /* parent on the upper right */
          if (ix < snb-1) {
            m += 4;

            shift_m2p(coeffs1+2*P,coeffs2+2*P,ucoeffs1+8*P,ucoeffs2+8*P,4,4);
            shift_m2p(coeffs1+3*P,coeffs2+3*P,ucoeffs1+9*P,ucoeffs2+9*P,4,1);
            shift_m2p(coeffs1+3*P,coeffs2+3*P,ucoeffs1+10*P,ucoeffs2+10*P,30,2);
          }

          /* parent above */
          shift_m2p(coeffs1,coeffs2,ucoeffs1+offset*P,ucoeffs2+offset*P,
          begin0,m);
          shift_m2p(coeffs1+P,coeffs2+P,ucoeffs1+offset*P,ucoeffs2+offset*P,
          begin1,m);
          shift_m2p(coeffs1+2*P,coeffs2+2*P,ucoeffs1+6*P,ucoeffs2+6*P,
          12,2);
          shift_m2p(coeffs1+3*P,coeffs2+3*P,ucoeffs1+6*P,ucoeffs2+6*P,
          26,2);
        }

        /* contribution from parent */
        shift_p2p(coeffs2,parent->coeff2);

        /* next parent */
        parent++;
        coeffs1 += 4*P;	ucoeffs1 += 4*P;
        coeffs2 += 4*P; ucoeffs2 += 4*P;
      }
  }
}
/*------------------------------------------------------------------------*/
void MPexp_smooth(int pot,SMOOTHER smooth,double xopt,bool cont,
		  double *cutoff,double *shape,double *scale)
/* Precomputes the three parameters (cutoff,shape,scale) given
   (pot,smooth,xopt,cutoff,cont). This yields a slicker code for
   mpexp_directInteract() below. */
{
  *scale = 1.0;

  if (pot == 0)
    switch (smooth) {
    case DIRAC:
      break;

    case RANKINE:
      *cutoff = xopt;
      *shape = log(*cutoff)-0.5;
      break;

    case SCULLY:
      *shape = xopt*xopt;
      if (cont)
        *scale = 2.0*log(*cutoff)/log(*shape+*cutoff**cutoff);
      break;

    case OSEEN:
      *shape = 1.2564312086261696770/(xopt*xopt);
      if (cont)
        *scale = log(*cutoff)/(log(*cutoff)+
        0.5*expint(*shape**cutoff**cutoff));
      break;
    }
  else
    switch (smooth) {
    case DIRAC:
      break;

    case RANKINE:
      *cutoff = xopt;
      *shape = 1.0/(*cutoff**cutoff);
      break;

    case SCULLY:
      *shape = xopt*xopt;
      if (cont)
        *scale = 1.0+*shape/(*cutoff**cutoff);
      break;

    case OSEEN:
      *shape = 1.2564312086261696770/(xopt*xopt);
      if (cont)
        *scale = -1.0/expm1(-*shape**cutoff**cutoff);
      break;
    }
}
/*------------------------------------------------------------------------*/
void MPexp_eval2(MPexp *this,
		 double *pr,double *pi,
		 const double *x,const double *y,
		 const double *mr,const double *mi,int pot,
		 SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Evaluation of the potential at the potentials themselves. */
{
  mpexp *offset = this->root+this->lptr[this->nlevel];
  const int snf = 1<<this->nlevel;
  double shape,scale;

  MPexp_smooth(pot,smooth,xopt,cont,&cutoff,&shape,&scale);
  for (int iy = 0,j = 0; iy < snf; iy++)
    for (int ix = 0; ix < snf; ix++,j++)
      if (this->ixptr[j] != this->ixptr[j+1]) {
	      /* contribution from distant potentials */
        mpexp_eval2(offset+j,this->pcoeff,pr,pi,x,y,
        this->ix,this->ixptr[j],this->ixptr[j+1]);

        /* symmetric interaction with at most 5 nearest neighbours:
           box immediate to the right */
        if (ix < snf-1)
          mpexp_directInteract2(pr,pi,x,y,mr,mi,this->ix,
          this->ixptr[j],this->ixptr[j+1],
          this->ixptr[j+1],this->ixptr[j+2],
          pot,smooth,cutoff,shape,scale);

	      /* row of boxes above */
        if (iy < snf-1) {
          const int begin = j+snf-(ix > 0),end = j+snf+1+(ix < snf-1);

          mpexp_directInteract2(pr,pi,x,y,mr,mi,this->ix,
          this->ixptr[j],this->ixptr[j+1],
          this->ixptr[begin],this->ixptr[end],
          pot,smooth,cutoff,shape,scale);
        }

	      /* direct interaction within box */
        mpexp_directInteract2(pr,pi,x,y,mr,mi,this->ix,
        this->ixptr[j],this->ixptr[j+1],
        -1,this->ixptr[j+1],
        pot,smooth,cutoff,shape,scale);
      }
}
/*------------------------------------------------------------------------*/
#define NULLPT(ptr,i) (ptr != NULL ? &ptr[i] : NULL)

void MPexp_eval(MPexp *this,
		double *qr,double *qi,
		int NE,const double *ex,const double *ey,
		const double *x,const double *y,
		const double *mr,const double *mi,int pot,
		SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Evaluation of the potential at the NE points (ex,ey). It is
   required that all points lie within the bounding box. */
{
  mpexp *offset = this->root+this->lptr[this->nlevel];
  const int snf = 1<<this->nlevel;
  const double hx = this->xmax-this->xmin; /* hy = hx by construction */
  double shape,scale;

  MPexp_smooth(pot,smooth,xopt,cont,&cutoff,&shape,&scale);
  for (int ii = 0; ii < NE; ii++) {
    /* determine the local box */
    const double xx = ex[ii],yy = ey[ii];
    const double lx = (xx-this->xmin)/hx,ly = (yy-this->ymin)/hx;
    const int ix = IMIN((int)floor(lx*snf),snf-1),
      iy = IMIN((int)floor(ly*snf),snf-1);
    const int j = ix+iy*snf;
    double *qii = NULLPT(qi,ii);

    /* contribution from distant potentials */
    mpexp_eval(offset+j,this->pcoeff,&qr[ii],qii,xx,yy);

    /* direct interaction with at most 9 boxes */
    const int begin = j-(ix > 0),end = j+1+(ix < snf-1);

    /* present row */
    mpexp_directInteract(&qr[ii],qii,xx,yy,
			 x,y,mr,mi,this->ix,
			 this->ixptr[begin],this->ixptr[end],
			 pot,smooth,cutoff,shape,scale);

    /* row below */
    if (iy > 0)
      mpexp_directInteract(&qr[ii],qii,xx,yy,
			   x,y,mr,mi,this->ix,
			   this->ixptr[begin-snf],this->ixptr[end-snf],
			   pot,smooth,cutoff,shape,scale);

    /* row above */
    if (iy < snf-1)
      mpexp_directInteract(&qr[ii],qii,xx,yy,
			   x,y,mr,mi,this->ix,
			   this->ixptr[begin+snf],this->ixptr[end+snf],
			   pot,smooth,cutoff,shape,scale);
  }
}
/*------------------------------------------------------------------------*/
void MPexp_free(MPexp *this)
/* Deallocates the whole tree. */
{
  mxFree(this->ixptr);
  mxFree(this->ix);
  mxFree(this->lptr);
  mxFree(this->root->coeff1);
  mxFree(this->root->coeff2);
  mxFree(this->root);
}
/*------------------------------------------------------------------------*/
void mpexp_init(mpexp *this,int p,
                const double x,const double y,
		const double mr,const double mi,int pot)
/* Initializes a multipole expansion at the lowest level with a single
   pointmass (mr,mi) at (x,y). */
{
  const dcmplx r = (x+I*y)-this->z0;
  const dcmplx m = mr+I*mi;
  dcmplx *coeff1 = this->coeff1;

  /* We have that the potential is given by (pot = 0)
     mp(z) = -m*log(z-zi)
           = -m*log(z-z0-(zi-z0))
	   = -m*log((z-z0)*(1-(zi-z0)/(z-z0)))
	   = -m*[log(z-z0)+log(1-(zi-z0)/(z-z0))]
	   = -m*[log(z-z0)-sum_{j >= 1} (zi-z0)^j/j*(z-z0)^{-j},
     or othwerwise by (pot = 1)
     mp(z) = -m*1/(z-zi)
           = -m*1/(z-z0-(zi-z0))
	   = -m/(z-z0)*1/(1-(zi-z0)/(z-z0))
	   = -m*sum_{j >= 1} (zi-z0)^{j-1}*(z-z0)^{-j}.
  */
  if (pot == 0) {
    dcmplx z = r*m;
    coeff1[0] -= m;
    for (int j = 1; j <= p; j++,z *= r) coeff1[j] += z/j;
  }
  else {
    dcmplx z = m;
    for (int j = 1; j <= p; j++,z *= r) coeff1[j] -= z;
  }
}
/*------------------------------------------------------------------------*/
void mpexp_eval2(mpexp *this,int p,
		 double *pr,double *pi,
		 const double *x,const double *y,
		 const int *ix,int begin,int end)
/* Evaluation of potential due to distant potentials at the potentials
   themselves. */
{
  const dcmplx *coeff = this->coeff2;

  for (int i = begin; i < end; i++) {
    dcmplx polyval = coeff[p];
    const dcmplx r = (x[ix[i]]+I*y[ix[i]])-this->z0;

    /* Horner's method */
    for (int j = p-1; j >= 0; j--) polyval = r*polyval+coeff[j];

    /* add the result */
    pr[ix[i]] += creal(polyval);
    if (pi != NULL) pi[ix[i]] += cimag(polyval);
  }
}
/*------------------------------------------------------------------------*/
void mpexp_eval(mpexp *this,int p,
                double *qr,double *qi,
		double ex,double ey)
/* Evaluation of potential due to distant potentials at a single point
   (ex,ey). */
{
  const dcmplx *coeff = this->coeff2;
  dcmplx polyval = coeff[p];
  const dcmplx r = (ex+I*ey)-this->z0;

  /* Horner's method */
  for (int j = p-1; j >= 0; j--) polyval = r*polyval+coeff[j];

  /* add the result */
  *qr += creal(polyval);
  if (qi != NULL) *qi += cimag(polyval);
}
/*------------------------------------------------------------------------*/
#define BEGIN(begin,i) (begin == -1 ? i+1 : begin)

void mpexp_directInteract2(double *pr,double *pi,
			   const double *zr,const double *zi,
			   const double *mr,const double *mi,
			   const int *ix,
			   int begin1,int end1,int begin2,int end2,
			   int pot,SMOOTHER smooth,
			   double cutoff,double shape,double scale)
/* Direct interaction between potentials inside a box (begin2 == -1)
   or between two boxes. */
{
  if (pot == 0) { /* logaritmic potential */
    if (mi == NULL) { /* rlog2() */
      switch (smooth) {
      case DIRAC:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const double P = log(hypot(zr[ix[i]]-zr[ix[j]],
				       zi[ix[i]]-zi[ix[j]]));
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	  }
	break;

      case RANKINE:
	for (int i = begin1; i < end1; i++) {
	  if (begin2 == -1)
	    pr[ix[i]] -= mr[ix[i]]*shape;
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const double ad = hypot(zr[ix[i]]-zr[ix[j]],zi[ix[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else
	      P = 0.5*(ad/cutoff)*(ad/cutoff)+shape;
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	  }
	}
	break;

      case SCULLY:
	for (int i = begin1; i < end1; i++) {
	  if (begin2 == -1)
	    pr[ix[i]] -= mr[ix[i]]*0.5*log(shape)*scale;
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const double ad = hypot(zr[ix[i]]-zr[ix[j]],zi[ix[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else
	      P = 0.5*log(shape+ad*ad)*scale;
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	  }
	}
	break;

      case OSEEN:
	for (int i = begin1; i < end1; i++) {
	  if (begin2 == -1)
	    pr[ix[i]] += mr[ix[i]]*0.5*(log(shape)+gamma_)*scale;
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const double ad = hypot(zr[ix[i]]-zr[ix[j]],zi[ix[i]]-zi[ix[j]]);
	    dcmplx P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else if (ad != 0.0)
	      P = (log(ad)+0.5*expint(shape*ad*ad))*scale;
	    else
	      P = -0.5*(log(shape)+gamma_)*scale;
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	  }
	}
	break;
      }
    }
    else { /* zlog2() */
      switch (smooth) {
      case DIRAC:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const double P = log(hypot(zr[ix[i]]-zr[ix[j]],
				       zi[ix[i]]-zi[ix[j]]));
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pi[ix[i]] -= mi[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	    pi[ix[j]] -= mi[ix[i]]*P;
	  }
	break;

      case RANKINE:
	for (int i = begin1; i < end1; i++) {
	  if (begin2 == -1) {
	    pr[ix[i]] -= mr[ix[i]]*shape;
	    pi[ix[i]] -= mi[ix[i]]*shape;
	  }
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const double ad = hypot(zr[ix[i]]-zr[ix[j]],zi[ix[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else
	      P = 0.5*(ad/cutoff)*(ad/cutoff)+shape;
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pi[ix[i]] -= mi[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	    pi[ix[j]] -= mi[ix[i]]*P;
	  }
	}
	break;

      case SCULLY:
	for (int i = begin1; i < end1; i++) {
	  if (begin2 == -1) {
	    pr[ix[i]] -= mr[ix[i]]*0.5*log(shape)*scale;
	    pi[ix[i]] -= mi[ix[i]]*0.5*log(shape)*scale;
	  }
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const double ad = hypot(zr[ix[i]]-zr[ix[j]],zi[ix[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else
	      P = 0.5*log(shape+ad*ad)*scale;
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pi[ix[i]] -= mi[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	    pi[ix[j]] -= mi[ix[i]]*P;
	  }
	}
	break;

      case OSEEN:
	for (int i = begin1; i < end1; i++) {
	  if (begin2 == -1) {
	    pr[ix[i]] += mr[ix[i]]*0.5*(log(shape)+gamma_)*scale;
	    pi[ix[i]] += mi[ix[i]]*0.5*(log(shape)+gamma_)*scale;
	  }
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const double ad = hypot(zr[ix[i]]-zr[ix[j]],zi[ix[i]]-zi[ix[j]]);
	    dcmplx P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else if (ad != 0.0)
	      P = (log(ad)+0.5*expint(shape*ad*ad))*scale;
	    else
	      P = -0.5*(log(shape)+gamma_)*scale;
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pi[ix[i]] -= mi[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	    pi[ix[j]] -= mi[ix[i]]*P;
	  }
	}
	break;
      }
    }
  }
  else { /* harmonic potential */
    if (mi == NULL) { /* rinv2() */
      switch (smooth) {
      case DIRAC:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const dcmplx P = 1.0/(zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]));
	    pr[ix[i]] -= mr[ix[j]]*creal(P);
	    pi[ix[i]] -= mr[ix[j]]*cimag(P);
	    pr[ix[j]] += mr[ix[i]]*creal(P);
	    pi[ix[j]] += mr[ix[i]]*cimag(P);
	  }
	break;

      case RANKINE:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const dcmplx d = zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    dcmplx P;
	    if (ad >= cutoff)
	      P = 1.0/d;
	    else
	      P = conj(d)*shape;
	    pr[ix[i]] -= mr[ix[j]]*creal(P);
	    pi[ix[i]] -= mr[ix[j]]*cimag(P);
	    pr[ix[j]] += mr[ix[i]]*creal(P);
	    pi[ix[j]] += mr[ix[i]]*cimag(P);
	  }
	break;

      case SCULLY:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const dcmplx d = zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    dcmplx P;
	    if (ad >= cutoff)
	      P = 1.0/d;
	    else
	      P = conj(d)/(shape+ad*ad)*scale;
	    pr[ix[i]] -= mr[ix[j]]*creal(P);
	    pi[ix[i]] -= mr[ix[j]]*cimag(P);
	    pr[ix[j]] += mr[ix[i]]*creal(P);
	    pi[ix[j]] += mr[ix[i]]*cimag(P);
	  }
	break;

      case OSEEN:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const dcmplx d = zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    if (ad != 0.0) {
	      dcmplx P;
	      if (ad >= cutoff)
		P = 1.0/d;
	      else
		P = -expm1(-shape*ad*ad)/d*scale;
	      pr[ix[i]] -= mr[ix[j]]*creal(P);
	      pi[ix[i]] -= mr[ix[j]]*cimag(P);
	      pr[ix[j]] += mr[ix[i]]*creal(P);
	      pi[ix[j]] += mr[ix[i]]*cimag(P);
	    }
	  }
	break;
      }
    }
    else { /* zinv2() */
      switch (smooth) {
      case DIRAC:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const dcmplx P = 1.0/(zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]));
	    pr[ix[i]] -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	    pi[ix[i]] -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
	    pr[ix[j]] += mr[ix[i]]*creal(P)-mi[ix[i]]*cimag(P);
	    pi[ix[j]] += mr[ix[i]]*cimag(P)+mi[ix[i]]*creal(P);
	  }
	break;

      case RANKINE:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const dcmplx d = zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    dcmplx P;
	    if (ad >= cutoff)
	      P = 1.0/d;
	    else
	      P = conj(d)*shape;
	    pr[ix[i]] -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	    pi[ix[i]] -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
	    pr[ix[j]] += mr[ix[i]]*creal(P)-mi[ix[i]]*cimag(P);
	    pi[ix[j]] += mr[ix[i]]*cimag(P)+mi[ix[i]]*creal(P);
	  }
	break;

      case SCULLY:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const dcmplx d = zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    dcmplx P;
	    if (ad >= cutoff)
	      P = 1.0/d;
	    else
	      P = conj(d)/(shape+ad*ad)*scale;
	    pr[ix[i]] -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	    pi[ix[i]] -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
	    pr[ix[j]] += mr[ix[i]]*creal(P)-mi[ix[i]]*cimag(P);
	    pi[ix[j]] += mr[ix[i]]*cimag(P)+mi[ix[i]]*creal(P);
	  }
	break;

      case OSEEN:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
	    const dcmplx d = zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    if (ad != 0.0) {
	      dcmplx P;
	      if (ad >= cutoff)
		P = 1.0/d;
	      else
		P = -expm1(-shape*ad*ad)/d*scale;
	      pr[ix[i]] -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	      pi[ix[i]] -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
	      pr[ix[j]] += mr[ix[i]]*creal(P)-mi[ix[i]]*cimag(P);
	      pi[ix[j]] += mr[ix[i]]*cimag(P)+mi[ix[i]]*creal(P);
	    }
	  }
	break;
      }
    }
  }
}
/*------------------------------------------------------------------------*/
void mpexp_directInteract(double *qr,double *qi,
			  double er,double ei,
			  const double *zr,const double *zi,
			  const double *mr,const double *mi,
			  const int *ix,int begin,int end,
			  int pot,SMOOTHER smooth,
			  double cutoff,double shape,double scale)
/* Direct interaction from potentials inside boxes computed at a
   single point (er,ei). */
{
  if (pot == 0) { /* logarithmic potential */
    if (mi == NULL) { /* rlog() */
      switch (smooth) {
      case DIRAC:
	for (int j = begin; j < end; j++)
	  if (er != zr[ix[j]] || ei != zi[ix[j]]) {
	    const double P = log(hypot(er-zr[ix[j]],ei-zi[ix[j]]));
	    *qr -= mr[ix[j]]*P;
	  }
	break;

      case RANKINE:
	for (int j = begin; j < end; j++) {
	  const double ad = hypot(er-zr[ix[j]],ei-zi[ix[j]]);
	  double P;
	  if (ad >= cutoff)
	    P = log(ad);
	  else
	    P = 0.5*(ad/cutoff)*(ad/cutoff)+shape;
	  *qr -= mr[ix[j]]*P;
	}
	break;

      case SCULLY:
	for (int j = begin; j < end; j++) {
	  const double ad = hypot(er-zr[ix[j]],ei-zi[ix[j]]);
	  double P;
	  if (ad >= cutoff)
	    P = log(ad);
	  else
	    P = 0.5*log(shape+ad*ad)*scale;
	  *qr -= mr[ix[j]]*P;
	}
	break;

      case OSEEN:
	for (int j = begin; j < end; j++) {
	  const double ad = hypot(er-zr[ix[j]],ei-zi[ix[j]]);
	  dcmplx P;
	  if (ad >= cutoff)
	    P = log(ad);
	  else if (ad != 0.0)
	    P = (log(ad)+0.5*expint(shape*ad*ad))*scale;
	  else
	    P = -0.5*(log(shape)+gamma_)*scale;
	  *qr -= mr[ix[j]]*P;
	}
	break;
      }
    }
    else { /* zlog() */
      switch (smooth) {
      case DIRAC:
	for (int j = begin; j < end; j++)
	  if (er != zr[ix[j]] || ei != zi[ix[j]]) {
	    const double P = log(hypot(er-zr[ix[j]],ei-zi[ix[j]]));
	    *qr -= mr[ix[j]]*P;
	    *qi -= mi[ix[j]]*P;
	  }
	break;

      case RANKINE:
	for (int j = begin; j < end; j++) {
	  const double ad = hypot(er-zr[ix[j]],ei-zi[ix[j]]);
	  double P;
	  if (ad >= cutoff)
	    P = log(ad);
	  else
	    P = 0.5*(ad/cutoff)*(ad/cutoff)+shape;
	  *qr -= mr[ix[j]]*P;
	  *qi -= mi[ix[j]]*P;
	}
	break;

      case SCULLY:
	for (int j = begin; j < end; j++) {
	  const double ad = hypot(er-zr[ix[j]],ei-zi[ix[j]]);
	  double P;
	  if (ad >= cutoff)
	    P = log(ad);
	  else
	    P = 0.5*log(shape+ad*ad)*scale;
	  *qr -= mr[ix[j]]*P;
	  *qi -= mi[ix[j]]*P;
	}
	break;

      case OSEEN:
	for (int j = begin; j < end; j++) {
	  const double ad = hypot(er-zr[ix[j]],ei-zi[ix[j]]);
	  dcmplx P;
	  if (ad >= cutoff)
	    P = log(ad);
	  else if (ad != 0.0)
	    P = (log(ad)+0.5*expint(shape*ad*ad))*scale;
	  else
	    P = -0.5*(log(shape)+gamma_)*scale;
	  *qr -= mr[ix[j]]*P;
	  *qr -= mi[ix[j]]*P;
	}
	break;
      }
    }
  }
  else { /* harmonic potential */
    if (mi == NULL) { /* rinv() */
      switch (smooth) {
      case DIRAC:
	for (int j = begin; j < end; j++)
	  if (er != zr[ix[j]] || ei != zi[ix[j]]) {
	    const dcmplx P = 1.0/(er-zr[ix[j]]+I*(ei-zi[ix[j]]));
	    *qr -= mr[ix[j]]*creal(P);
	    *qi -= mr[ix[j]]*cimag(P);
	  }
	break;

      case RANKINE:
	for (int j = begin; j < end; j++) {
	  const dcmplx d = er-zr[ix[j]]+I*(ei-zi[ix[j]]);
	  const double ad = cabs(d);
	  dcmplx P;
	  if (ad >= cutoff)
	    P = 1.0/d;
	  else
	    P = conj(d)*shape;
	  *qr -= mr[ix[j]]*creal(P);
	  *qi -= mr[ix[j]]*cimag(P);
	}
	break;

      case SCULLY:
	for (int j = begin; j < end; j++) {
	  const dcmplx d = er-zr[ix[j]]+I*(ei-zi[ix[j]]);
	  const double ad = cabs(d);
	  dcmplx P;
	  if (ad >= cutoff)
	    P = 1.0/d;
	  else
	    P = conj(d)/(shape+ad*ad)*scale;
	  *qr -= mr[ix[j]]*creal(P);
	  *qi -= mr[ix[j]]*cimag(P);
	}
	break;

      case OSEEN:
	for (int j = begin; j < end; j++) {
	  const dcmplx d = er-zr[ix[j]]+I*(ei-zi[ix[j]]);
	  const double ad = cabs(d);
	  if (ad != 0.0) {
	    dcmplx P;
	    if (ad >= cutoff)
	      P = 1.0/d;
	    else
	      P = -expm1(-shape*ad*ad)/d*scale;
	    *qr -= mr[ix[j]]*creal(P);
	    *qi -= mr[ix[j]]*cimag(P);
	  }
	}
	break;
      }
    }
    else { /* zinv() */
      switch (smooth) {
      case DIRAC:
	for (int j = begin; j < end; j++)
	  if (er != zr[ix[j]] || ei != zi[ix[j]]) {
	    const dcmplx P = 1.0/(er-zr[ix[j]]+I*(ei-zi[ix[j]]));
	    *qr -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	    *qi -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
	  }
	break;

      case RANKINE:
	for (int j = begin; j < end; j++) {
	  const dcmplx d = er-zr[ix[j]]+I*(ei-zi[ix[j]]);
	  const double ad = cabs(d);
	  dcmplx P;
	  if (ad >= cutoff)
	    P = 1.0/d;
	  else
	    P = conj(d)*shape;
	  *qr -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	  *qi -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
	}
	break;

      case SCULLY:
	for (int j = begin; j < end; j++) {
	  const dcmplx d = er-zr[ix[j]]+I*(ei-zi[ix[j]]);
	  const double ad = cabs(d);
	  dcmplx P;
	  if (ad >= cutoff)
	    P = 1.0/d;
	  else
	    P = conj(d)/(shape+ad*ad)*scale;
	  *qr -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	  *qi -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
	}
	break;

      case OSEEN:
	for (int j = begin; j < end; j++) {
	  const dcmplx d = er-zr[ix[j]]+I*(ei-zi[ix[j]]);
	  const double ad = cabs(d);
	  if (ad != 0.0) {
	    dcmplx P;
	    if (ad >= cutoff)
	      P = 1.0/d;
	    else
	      P = -expm1(-shape*ad*ad)/d*scale;
	    *qr -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	    *qi -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
	  }
	}
	break;
      }
    }
  }
}
/*------------------------------------------------------------------------*/

#endif /* __fmm_h */
