/* fmm2du.c -- Fast uniform multipole method for 2-D potentials in free space */

/* S. Engblom 2007-07-08 (Major revision) */
/* S. Engblom 2007-05-14 (Revision) */
/* S. Engblom 2006-10-11 (Revision continued) */
/* S. Engblom 2006-07-10 (Port to Mex and a major revision) */
/* S. Engblom 2005-01-05 */

#include <math.h>

#ifndef NO_COMPLEX
#include <complex.h>
#else
/* fix for CC under Solaris 9 (non-conforming C99-math) */
#define complex _Complex
#define I _Complex_I

#define creal(z) (((double *) (&z))[0])
#define cimag(z) (((double *) (&z))[1])

double complex clog(double complex z)
{
  double re = creal(z),im = cimag(z);
  return log(hypot(re,im))+I*atan2(im,re);
}

double cabs(double complex z)
{
  return hypot(creal(z),cimag(z));
}

double complex conj(double complex z)
{
  return creal(z)-I*cimag(z);
}

double log2(double x)
{
  return log(x)/0.69314718055994530942;
}
#endif

#include <float.h>
#include <string.h>

#include "mex.h"
#include "matrix.h"

/* forward declarations */
typedef enum {POT = 8,TOL = 3,NDIRECT = 7,SMOOTH = 4,CUTOFF = 5,
	      XOPT = 1,CONT = 2,ERR_PROPERTY = -1} PROPERTY;
typedef enum {DIRAC = 3,RANKINE = 1,SCULLY = 0,OSEEN = 2,
	      ERR_SMOOTHER = -1} SMOOTHER;
typedef void (*getvalFun)(void *,const mxArray *);

void getXOPT(void *,const mxArray *);
void getCONT(void *,const mxArray *);
void getTOL(void *,const mxArray *);
void getSMOOTH(void *,const mxArray *);
void getCUTOFF(void *,const mxArray *);
void getNDIRECT(void *,const mxArray *);
void getPOT(void *,const mxArray *);

PROPERTY getprop(const char *str);
SMOOTHER getsmooth(const char *str);
double expint(double x); /*** temporary ***/

/* declaration of public interfaces */
void directInteract(int N,
                    const double *zr,const double *zi,
		    const double *mr,const double *mi,
		    int NE,
		    const double *er,const double *ei,
                    double *pr,double *pi,
		    double *qr,double *qi,int pot,
		    SMOOTHER smooth,double xopt,double cutoff,bool cont);
void fmm2dInteract(int N,
		   const double *zr,const double *zi,
		   const double *mr,const double *mi,
		   int NE,
		   const double *er,const double *ei,
		   double *pr,double *pi,
		   double *qr,double *qi,
		   int pot,double tol,int Ndirect,
		   SMOOTHER smooth,double xopt,double cutoff,bool cont);

typedef double complex dcmplx;

const double pi_ = 3.1415926535897932384626434;
const double gamma_ = 0.5772156649015328606065121;

#define ISDOUBLEMATRIX(A) (mxIsDouble(A) && !mxIsSparse(A) && \
			   mxGetNumberOfDimensions(A) == 2)
#define ISREALSCALAR(A) (!mxIsCell(A) && !mxIsStruct(A) && \
			 !mxIsComplex(A) && mxGetNumberOfElements(A) == 1)
#define ISLOGICALSCALAR(A) (mxIsLogical(A) && \
			    mxGetNumberOfElements(A) == 1)

/* inclusion of source */
#include "expint.h"
#include "direct.h"
#include "fmm.h"

/*------------------------------------------------------------------------*/
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{
  /* main input */
  const double *zr,*zi,*mr,*mi,*er,*ei;
  int N,NE;

  /* values of properties with defaults */
  double xopt = 1e-3,cutoff = 2e-3,tol = 1e-6;
  int pot = 0,Ndirect = 5;
  SMOOTHER smooth = DIRAC;
  bool cont = true;

  /* output */
  double *pr = NULL,*pi = NULL,*qr = NULL,*qi = NULL;
  int eval_syntax = 2 < nrhs && ISDOUBLEMATRIX(prhs[2]);
  mxComplexity kind;

  /* parsing functions */
  static const getvalFun getval[] = {NULL,&getXOPT,&getCONT,&getTOL,
				     &getSMOOTH,&getCUTOFF,NULL,&getNDIRECT,
				     &getPOT};
  void *val[] = {NULL,&xopt,&cont,&tol,&smooth,&cutoff,NULL,&Ndirect,&pot};

  /*** temporary: test of expint() ***/
  if (nrhs == 1) {
    double *p = mxGetPr(plhs[0] = mxDuplicateArray(prhs[0]));
    for (int i = 0; i < mxGetNumberOfElements(prhs[0]); i++)
      p[i] = expint(p[i]);
    return;
  }

 /* syntax */
  if (2 < nlhs || nrhs < 2 || (nrhs-eval_syntax)&1)
    mexErrMsgTxt("Expecting one or two outputs and two or three inputs "
		 "followed by property/value-pairs.");
  if (nlhs > 1 && !eval_syntax)
    mexErrMsgTxt("Three matrices required for two outputs.");

  /* inputs Z and M */
  if (!ISDOUBLEMATRIX(prhs[0]) || !ISDOUBLEMATRIX(prhs[1]) ||
      mxGetM(prhs[0]) != mxGetM(prhs[1]) || 
      mxGetN(prhs[0]) != mxGetN(prhs[1]))
    mexErrMsgTxt("First two arguments must be double matrices of " 
		 "matching sizes.");
  N = mxGetNumberOfElements(prhs[0]);
  if (N != 0 && !mxIsComplex(prhs[0]))
    mexErrMsgTxt("First argument must be complex.");

  zr = mxGetPr(prhs[0]); zi = mxGetPi(prhs[0]);
  mr = mxGetPr(prhs[1]); mi = mxGetPi(prhs[1]);

  /* input ZE -- or just (er,ei) for short */
  if (eval_syntax) {
    NE = mxGetNumberOfElements(prhs[2]);
    if (NE != 0 && !mxIsComplex(prhs[2]))
      mexErrMsgTxt("Evaluation points must be complex.");
    er = mxGetPr(prhs[2]); ei = mxGetPi(prhs[2]);
  }

  /* parse property/value-pairs in an automatic fashion */
  for (int i = 2+eval_syntax; i < nrhs-1; i += 2) {
    PROPERTY prop;
    char str[10];

    /* determine property */
    if (!mxIsChar(prhs[i]))
      mexErrMsgTxt("Expecting a property name.");
    if (mxGetString(prhs[i],str,10) != 0 ||
	(prop = getprop(str)) == ERR_PROPERTY)
      mexErrMsgTxt("Unknown property.");

    /* get its value */
    getval[prop](val[prop],prhs[i+1]);
  }

  /* remaining check */
  if (cutoff <= xopt)
    mexErrMsgTxt("Parameter 'cutoff' must be larger than 'xopt'.");

  /* complexity */
  kind = pot == 1 || mi != NULL ? mxCOMPLEX : mxREAL;

  /* allocate and produce output */
  if (nlhs > 1) {
    /* evaluate both at potentials Z and at coordinates ZE */
    plhs[0] = mxCreateDoubleMatrix(mxGetM(prhs[0]),mxGetN(prhs[0]),kind);
    pr = mxGetPr(plhs[0]); pi = mxGetPi(plhs[0]);

    plhs[1] = mxCreateDoubleMatrix(mxGetM(prhs[2]),mxGetN(prhs[2]),kind);
    qr = mxGetPr(plhs[1]); qi = mxGetPi(plhs[1]);
  }
  else if (!eval_syntax || zr == er && zi == ei && N == NE) {
    /* only at potentials Z */
    plhs[0] = mxCreateDoubleMatrix(mxGetM(prhs[0]),mxGetN(prhs[0]),kind);
    pr = mxGetPr(plhs[0]); pi = mxGetPi(plhs[0]);
  }
  else {
    /* at coordinates ZE */
    plhs[0] = mxCreateDoubleMatrix(mxGetM(prhs[2]),mxGetN(prhs[2]),kind);
    qr = mxGetPr(plhs[0]); qi = mxGetPi(plhs[0]);
  }

  /* evaluate */
  if (tol == 0.0)
    directInteract(N,zr,zi,mr,mi,NE,er,ei,pr,pi,qr,qi,
		   pot,smooth,xopt,cutoff,cont);
  else
    fmm2dInteract(N,zr,zi,mr,mi,NE,er,ei,pr,pi,qr,qi,
		  pot,tol,Ndirect,smooth,xopt,cutoff,cont);
}
/*------------------------------------------------------------------------*/
PROPERTY getprop(const char *str)
/* Small perfect hash for the supported properties. Note that str[1]
   and str[2] must be readable for this to work. Returns ERR_PROPERTY
   on failure. */
{
  static const char *tab[] = {"","xopt","cont","tol","smooth","cutoff",
			      "","ndirect","pot"};
  const int hash = ((int)str[1]+(int)4*str[2])%9;

  return str[0] != '\0' && strcmp(str,tab[hash]) == 0 ? 
    hash : ERR_PROPERTY;
}
/*------------------------------------------------------------------------*/
void getXOPT(void *xopt_,const mxArray *rhs)
{
  double *xopt = (double *)xopt_;

  if (!ISREALSCALAR(rhs))
      mexErrMsgTxt("Expecting a real and scalar option 'xopt'.");
  *xopt = *mxGetPr(rhs);
  if (*xopt <= 0)
    mexErrMsgTxt("Property 'xopt' must be strictly positive.");
}
/*------------------------------------------------------------------------*/
void getCONT(void *cont_,const mxArray *rhs)
{
  bool *cont = (bool *)cont_;

  if (!ISLOGICALSCALAR(rhs))
    mexErrMsgTxt("Expecting a logical scalar option 'cont'.");
  *cont = mxIsLogicalScalarTrue(rhs);
}
/*------------------------------------------------------------------------*/
void getTOL(void *tol_,const mxArray *rhs)
{
  double *tol = (double *)tol_;

  if (!ISREALSCALAR(rhs))
      mexErrMsgTxt("Expecting a real and scalar option 'tol'.");
  *tol = *mxGetPr(rhs);
  if (*tol < 0)
    mexErrMsgTxt("Property 'tol' must not be negative.");
}
/*------------------------------------------------------------------------*/
void getSMOOTH(void *smooth_,const mxArray *rhs)
{
  SMOOTHER *smooth = (SMOOTHER *)smooth_;
  char str[10];

  if (!mxIsChar(rhs))
    mexErrMsgTxt("Expecting a character array as option 'smooth'.");
  if (mxGetString(rhs,str,10) != 0 || 
      (*smooth = getsmooth(str)) == ERR_SMOOTHER)
    mexErrMsgTxt("Unknown smoother.");
}
/*------------------------------------------------------------------------*/
void getCUTOFF(void *cutoff_,const mxArray *rhs)
{
  double *cutoff = (double *)cutoff_;

  if (!ISREALSCALAR(rhs))
      mexErrMsgTxt("Expecting a real and scalar option 'cutoff'.");
  *cutoff = *mxGetPr(rhs);
  if (*cutoff <= 0)
    mexErrMsgTxt("Property 'cutoff' must be strictly positive.");
}
/*------------------------------------------------------------------------*/
void getNDIRECT(void *Ndirect_,const mxArray *rhs)
{
  int *Ndirect = (int *)Ndirect_;

  if (!ISREALSCALAR(rhs))
    mexErrMsgTxt("Expecting a real and scalar option 'ndirect'.");
  *Ndirect = *mxGetPr(rhs);
  if (*Ndirect <= 0)
    mexErrMsgTxt("Property 'ndirect' must be strictly positive.");
}
/*------------------------------------------------------------------------*/
void getPOT(void *pot_,const mxArray *rhs)
{
  int *pot = (int *)pot_;

  if (!ISREALSCALAR(rhs))
    mexErrMsgTxt("Expecting a real and scalar option 'pot'.");
  *pot = *mxGetPr(rhs);
  if (*pot != 0 && *pot != 1)
    mexErrMsgTxt("Property 'pot' must be 0 or 1.");
}
/*------------------------------------------------------------------------*/
SMOOTHER getsmooth(const char *str)
/* Small perfect hash for the supported smoothers. Note that str[4]
   must be readable for this to work. Returns ERR_SMOOTHER on
   failure. */
{
  static const char *tab[] = {"scully","rankine","oseen","dirac"};
  const int hash = ((int)str[4])%4;

  return strcmp(str,tab[hash]) == 0 ? hash : ERR_SMOOTHER;
}
/*------------------------------------------------------------------------*/
