/* fmmshift.h */

/* S. Engblom 2007-05-29 */

#ifndef __fmmshift_h
#define __fmmshift_h

/* Definition of the three shift-operators used by the Fast Multipole
   Method. Performance is critical in general and in particular for
   the m2p-shift. */

/* Note: All integer arguments in calls to the BLAS functions will be
   casted according to (int *)&BLASINT. This is to provide for the
   (bad!) platforms where sizeof(int) < sizeof(size_t) and the binary
   library uses size_t while the declaration is erroneously int. For
   such cases you will have to manually define BLASINT as size_t when
   calling the compiler. */
#ifndef BLASINT
#define BLASINT int
#endif

void shift_alloc(int p,int pot);
void shift_initr(dcmplx *pow,dcmplx r);
void shift_initLevel(dcmplx r,bool downward);
void shift_atExit(void);

void shift_m2m(dcmplx *this,const dcmplx *that);
void shift_m2p(const dcmplx *this1,dcmplx *this2,
	       const dcmplx *that1,dcmplx *that2,int ix,int m);
void shift_p2p(dcmplx *this,const dcmplx *that);

/* declaration of BLAS-kernels */

/* C := alpha*op( A )*op( B ) + beta*C */
void dgemm_(const char *transA,const char *transB,
	    const int *M,const int *N,const int *K,
	    const double *alpha,const double *A,const int *ldA,
	    const double *B,const int *ldB,
	    const double *beta,
	    double *C,const int *ldC);

/* B := alpha*op( A )*B or B := alpha*B*op( A ) */
void dtrmm_(const char *side,const char *uplo,
	    const char *transA,const char *diag,
	    const int *M,const int *N,
	    const double *alpha,const double *A,const int *ldA,
	    double *B,const int *ldB);

/* precomputed data and workspace arrays */
static double *binomial1,*binomial2,*binomial3;
static dcmplx *rpow,*rtab,*logrtab;
static dcmplx *wksp,*wksp1,*wksp2;
static int pshift,potshift;

/*------------------------------------------------------------------------*/
void shift_alloc(int p,int pot)
/* Allocates and initializes the tables given a polynomial order p. */
{
  if (pshift != p) {
    /* tables of binomial coefficients */
    mxFree(binomial1);
    binomial1 = mxCalloc(p*p,sizeof(double));
    mexMakeMemoryPersistent(binomial1);
    mxFree(binomial2);
    binomial2 = mxCalloc(p*(p+1),sizeof(double));
    mexMakeMemoryPersistent(binomial2);
    mxFree(binomial3);
    binomial3 = mxCalloc((p+1)*(p+1),sizeof(double));
    mexMakeMemoryPersistent(binomial3);

    /* tables of powers of distances */
    mxFree(rpow);
    rpow = mxMalloc(4*(p+1)*sizeof(dcmplx));
    mexMakeMemoryPersistent(rpow);
    mxFree(rtab);
    rtab = mxMalloc(34*(p+1)*sizeof(dcmplx));
    mexMakeMemoryPersistent(rtab);
    mxFree(logrtab);
    logrtab = mxMalloc(34*2*sizeof(dcmplx));
    mexMakeMemoryPersistent(logrtab);

    /* workspace arrays */
    mxFree(wksp);
    wksp = mxMalloc(4*(p+1)*sizeof(dcmplx));
    mexMakeMemoryPersistent(wksp);
    mxFree(wksp1);
    wksp1 = mxMalloc(24*p*sizeof(dcmplx));
    mexMakeMemoryPersistent(wksp1);
    mxFree(wksp2);
    wksp2 = mxMalloc(24*(p+1)*sizeof(dcmplx));
    mexMakeMemoryPersistent(wksp2);

    pshift = p;
    potshift = pot;
    mexAtExit(shift_atExit);

    /* multipole to multipole shift */
    for (int l = 0; l < p; l++) {
      binomial1[p*l] = 1.0;
      for (int k = 1; k <= l; k++)
        binomial1[p*l+k] = binomial1[p*(l-1)+k]+binomial1[p*(l-1)+k-1];
    }

    /* multipole to polynomial shift */
    for (int l = 0; l <= p; l++)
      binomial2[p*l] = -1.0;
    for (int k = 1; k < p; k++) {
      binomial2[k] = -binomial2[k-1];
      for (int l = 1; l <= p; l++)
        binomial2[p*l+k] = binomial2[p*(l-1)+k]-binomial2[p*l+k-1];
    }

    /* polynomial to polynomial shift */
    binomial3[0] = 1.0;
    for (int k = 1; k <= p; k++) {
      binomial3[k] = -binomial3[k-1];
      for (int l = 1; l <= k; l++)
        binomial3[(p+1)*l+k] = binomial3[(p+1)*(l-1)+k-1]-
        binomial3[(p+1)*l+k-1];
    }
  }
}
/*------------------------------------------------------------------------*/
void shift_initr(dcmplx *pow,dcmplx r)
/* Initializes table with inverse powers of r. */
{
  pow[0] = 1.0;
  for (int k = 1; k <= pshift; k++)
    pow[k] = pow[k-1]/r;
}
/*------------------------------------------------------------------------*/
void shift_initLevel(dcmplx r,bool downward)
/* Initializes the table rpow with powers of r on a new level. If
   downward is true, then the table rtab containing inverse powers of
   several different (derived) r's is also initialized. */
{
  /* build powers of r, conj(-r), conj(r) and -r */
  if (rpow[0] != r) {
    dcmplx *pow0 = &rpow[0],*pow1 = &rpow[pshift+1],
      *pow2 = &rpow[2*(pshift+1)],*pow3 = &rpow[3*(pshift+1)];
    pow0[0] = pow1[0] = pow2[0] = pow3[0] = 1.0;
    for (int k = 1; k <= pshift; k++) {
      pow0[k] = pow0[k-1]*r;
      pow1[k] = pow1[k-1]*conj(-r);
      pow2[k] = pow2[k-1]*conj(r);
      pow3[k] = pow3[k-1]*(-r);
    }
  }

  /* tables containing powers and logarithms of distances to boxes in
     the interaction list*/
  if (downward) {
    const double hx = 2.0*creal(r),hy = 2.0*cimag(r);

    /* first part */
    shift_initr(&rtab[0*(pshift+1)],2.0*hx-I*hy);
    shift_initr(&rtab[1*(pshift+1)],3.0*hx-I*hy);
    shift_initr(&rtab[2*(pshift+1)],2.0*hx);
    shift_initr(&rtab[3*(pshift+1)],3.0*hx);
    shift_initr(&rtab[4*(pshift+1)],2.0*hx+I*hy);
    shift_initr(&rtab[5*(pshift+1)],3.0*hx+I*hy);
    shift_initr(&rtab[6*(pshift+1)],2.0*hx+I*2.0*hy);
    shift_initr(&rtab[7*(pshift+1)],3.0*hx+I*2.0*hy);

    /* second part */
    shift_initr(&rtab[8*(pshift+1)],-2.0*hx+I*2.0*hy);
    shift_initr(&rtab[9*(pshift+1)],-hx+I*2.0*hy);
    shift_initr(&rtab[10*(pshift+1)],-2.0*hx+I*3.0*hy);
    shift_initr(&rtab[11*(pshift+1)],-hx+I*3.0*hy);
    shift_initr(&rtab[12*(pshift+1)],I*2.0*hy);
    shift_initr(&rtab[13*(pshift+1)],hx+I*2.0*hy);
    shift_initr(&rtab[14*(pshift+1)],I*3.0*hy);
    shift_initr(&rtab[15*(pshift+1)],hx+I*3.0*hy);
    memcpy(&rtab[16*(pshift+1)],&rtab[6*(pshift+1)],2*(pshift+1)*sizeof(dcmplx));
    shift_initr(&rtab[18*(pshift+1)],2.0*hx+I*3.0*hy);
    shift_initr(&rtab[19*(pshift+1)],3.0*hx+I*3.0*hy);

    /* third part */
    shift_initr(&rtab[20*(pshift+1)],-3.0*hx+I*hy);
    shift_initr(&rtab[21*(pshift+1)],-2.0*hx+I*hy);
    shift_initr(&rtab[22*(pshift+1)],-3.0*hx+I*2.0*hy);
    memcpy(&rtab[23*(pshift+1)],&rtab[8*(pshift+1)],(pshift+1)*sizeof(dcmplx));
    shift_initr(&rtab[24*(pshift+1)],-3.0*hx+I*3.0*hy);
    memcpy(&rtab[25*(pshift+1)],&rtab[10*(pshift+1)],(pshift+1)*sizeof(dcmplx));
    memcpy(&rtab[26*(pshift+1)],&rtab[9*(pshift+1)],(pshift+1)*sizeof(dcmplx));
    memcpy(&rtab[27*(pshift+1)],&rtab[12*(pshift+1)],(pshift+1)*sizeof(dcmplx));
    memcpy(&rtab[28*(pshift+1)],&rtab[11*(pshift+1)],(pshift+1)*sizeof(dcmplx));
    memcpy(&rtab[29*(pshift+1)],&rtab[14*(pshift+1)],(pshift+1)*sizeof(dcmplx));
    memcpy(&rtab[30*(pshift+1)],&rtab[13*(pshift+1)],(pshift+1)*sizeof(dcmplx));
    memcpy(&rtab[31*(pshift+1)],&rtab[16*(pshift+1)],(pshift+1)*sizeof(dcmplx));
    memcpy(&rtab[32*(pshift+1)],&rtab[15*(pshift+1)],(pshift+1)*sizeof(dcmplx));
    memcpy(&rtab[33*(pshift+1)],&rtab[18*(pshift+1)],(pshift+1)*sizeof(dcmplx));

    /* logarithms */
    if (potshift == 0)
      for (int i = 0; i < 34; i++) {
        logrtab[2*i] = -clog(-rtab[i*(pshift+1)+1]);
        logrtab[2*i+1] = -clog(rtab[i*(pshift+1)+1]);
      }
  }
}
/*------------------------------------------------------------------------*/
void shift_atExit(void)
/* Deallocates persistent data. */
{
  mxFree(wksp2);
  mxFree(wksp1);
  mxFree(wksp);
  mxFree(logrtab);
  mxFree(rtab);
  mxFree(rpow);
  mxFree(binomial3);
  mxFree(binomial2);
  mxFree(binomial1);
  pshift = 0;
}
/*------------------------------------------------------------------------*/
void shift_m2m(dcmplx *this,const dcmplx *that)
/* Multipole to multipole shift from 4 children to 1 parent; this +=
   shift(that[0...p]) {child 0} + shift(that[p+1...2*p]) {child 1} +
   ...  and so on. */
{
  /* prescale */
  for (int i = 0,j = 1; i < 4; i++,j++) 
    for (int k = 0; k < pshift; k++,j++)
      wksp[4*k+i] = that[j]/rpow[j];

  /* multiplication with lower triangular Pascal matrix */
  const double one = 1.0;
  const char *si = "R",*ul = "U",*tr = "N",*di = "U";
  const BLASINT eight = 8,K = pshift;
  dtrmm_(si,ul,tr,di,(int *)&eight,(int *)&K,
	 &one,binomial1,(int *)&K,
	 (double *)wksp,(int *)&eight);

  /* postscale */
  if (potshift == 0)
    for (int i = 0,j = 0; i < 4; i++) {
      const dcmplx that0 = that[j];
      this[0] += that0;
      j++;
      for (int k = 0; k < pshift; k++,j++)
	this[k+1] += (wksp[4*k+i]-that0/(k+1))*rpow[j];
    }
  else
    for (int i = 0,j = 1; i < 4; i++,j++)
      for (int k = 0; k < pshift; k++,j++)
	this[k+1] += wksp[4*k+i]*rpow[j];
}
/*------------------------------------------------------------------------*/
void shift_m2p(const dcmplx *this1,dcmplx *this2,
	       const dcmplx *that1,dcmplx *that2,int ix,int m)
/* Blocked symmetric multipole to polynomial shift; this2 +=
   shift(that1) followed by that2 += shift(this1). ix is the starting
   index into rtab (and logrtab) and m is the number of consecutive
   shifts performec (that is, consecutive columns in that1/that2). */
{
  const dcmplx *pow = &rtab[ix*(pshift+1)];

  /* prescale */
  for (int i = 0,j = 1; i < m; i++,j++) {
    for (int k = 0; k < pshift>>1; k++) {
      wksp1[2*(m*2*k+i)] = that1[j]*pow[j];
      wksp1[2*(m*2*k+i)+1] = -this1[2*k+1]*pow[j];
      j++;
      wksp1[2*(m*(2*k+1)+i)] = that1[j]*pow[j];
      wksp1[2*(m*(2*k+1)+i)+1] = this1[2*k+2]*pow[j];
      j++;
    }
    if (pshift&1) {
      wksp1[2*(m*(pshift-1)+i)] = that1[j]*pow[j];
      wksp1[2*(m*(pshift-1)+i)+1] = -this1[pshift]*pow[j];
      j++;
    }
  }

  /* multiplication with rectangular Pascal matrix */
  const double one = 1.0,zero = 0.0;
  const char *tr = "N";
  const BLASINT M = 4*m,N = pshift+1,K = pshift;
  dgemm_(tr,tr,(int *)&M,(int *)&N,(int *)&K,
	 &one,(double *)wksp1,(int *)&M,
	 binomial2,(int *)&K,
	 &zero,(double *)wksp2,(int *)&M);

  /* postscale */
  if (potshift == 0)
    for (int i = 0,j = 0; i < m; i++) {
      const dcmplx that0 = that1[j];
      this2[0] += wksp2[2*i]+that0*logrtab[2*(ix+i)];
      that2[j] += wksp2[2*i+1]+this1[0]*logrtab[2*(ix+i)+1];
      j++;
      for (int k = 1; k <= pshift>>1; k++) {
	this2[2*k-1] += (wksp2[2*(m*(2*k-1)+i)]-that0/(2*k-1))*pow[j];
	that2[j] -= (wksp2[2*(m*(2*k-1)+i)+1]-this1[0]/(2*k-1))*pow[j];
	j++;
	this2[2*k] += (wksp2[2*(m*2*k+i)]-that0/(2*k))*pow[j];
	that2[j] += (wksp2[2*(m*2*k+i)+1]-this1[0]/(2*k))*pow[j];
	j++;
      }
      if (pshift&1) {
	this2[pshift] += (wksp2[2*(m*pshift+i)]-that0/pshift)*pow[j];
	that2[j] -= (wksp2[2*(m*pshift+i)+1]-this1[0]/pshift)*pow[j];
	j++;
      }
    }
  else
    for (int i = 0,j = 0; i < m; i++) {
      for (int k = 0; k <= pshift-1>>1; k++) {
	this2[2*k] += wksp2[2*(m*2*k+i)]*pow[j];
	that2[j] += wksp2[2*(m*2*k+i)+1]*pow[j];
	j++;
	this2[2*k+1] += wksp2[2*(m*(2*k+1)+i)]*pow[j];
	that2[j] -= wksp2[2*(m*(2*k+1)+i)+1]*pow[j];
	j++;
      }
      if (~pshift&1) {
	this2[pshift] += wksp2[2*(m*pshift+i)]*pow[j];
	that2[j] += wksp2[2*(m*pshift+i)+1]*pow[j];
	j++;
      }
    }
}
/*------------------------------------------------------------------------*/
void shift_p2p(dcmplx *this,const dcmplx *that)
/* Polynomial to polynomial shift from 1 parent to 4 children;
   this[0..p] += shift(that) {child 0}, this[p+1..2*p] += shift(that)
   {child 1}, and so on. */
{
  /* prescale */
  for (int i = 0,j = 0; i < 4; i++)
    for (int k = 0; k <= pshift; k++,j++)
      wksp[4*k+i] = that[k]*rpow[j];

  /* multiplication with upper triangular Pascal matrix */
  const double one = 1.0;
  const char *si = "R",*ul = "L",*tr = "N",*di = "U";
  const BLASINT eight = 8,N = pshift+1;
  dtrmm_(si,ul,tr,di,(int *)&eight,(int *)&N,
	 &one,binomial3,(int *)&N,
	 (double *)wksp,(int *)&eight);

  /* postscale */
  for (int i = 0,j = 0; i < 4; i++)
    for (int k = 0; k <= pshift; k++,j++)
      this[j] += wksp[4*k+i]/rpow[j];
}
/*------------------------------------------------------------------------*/
#endif /* __fmmshift_h */
